/* $Id: beacon.cpp 444 2011-02-20 22:11:04Z oh2gve $
 * 
 * Copyright 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Build config.
#ifdef __WXMSW__
#	include "windows/config.h"
#endif
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

// own stuff.
#include "beacon.h"
#include "cfg.h"
#include "dupefilter.h"
#include "connection.h"

// std stuff.
#include <string>
using std::string;
#include <cmath>
#include <sstream>
using std::ostringstream;
#include <vector>
using std::vector;
#include <fstream>
using std::ifstream;
#include <map>
using std::multimap;
using std::make_pair;
#include <ctime>
#include <list>
using std::list;

// wx stuff.
#include <wx/thread.h>
#include <wx/log.h>
#include <wx/string.h>



Beacon::Beacon(Cfg::Configuration const& cfg, DupeFilter* dupefilter):
wxThread(wxTHREAD_JOINABLE),
mutex_(), initialized_(false), keepgoing_(true),
dupefilter_(dupefilter), beaconBody_(), owncall_(cfg.callsign.ToAscii()), myInterval_(cfg.beaconInterval*60),
objFileName_(cfg.objectFile.ToAscii()), objects_()
{
	// Choose symbol.
	char symboltable = 'R';
	if ( cfg.uplink.serverState != Cfg::WRITEONLY )
	{
		symboltable = 'I';
	}
	char symbolcode = '&';
	
	// Encode coordinates.
	double latval = 380926 * (90 - cfg.location.lat);
	double lonval = 190463 * (180 + cfg.location.lon);
	string latstring = "";
	string lonstring = "";
	for ( int i = 3; i >= 0; i-- )
	{
		// Latitude branch.
		long value = latval / pow(91, i);
		latval = fmod(latval, pow(91, i));
		latstring.append(1, value+33);
		
		// Longitude branch.
		value = lonval / pow(91, i);
		lonval = fmod(lonval, pow(91, i));
		lonstring.append(1, value+33);
	}
	
	// Construct packet body.
	ostringstream body;
	body << "!"; // no messaging, we're just a stupid automatic APRS node
	body << symboltable << latstring << lonstring << symbolcode;
	body << " sT"; // no course, speed or type
	body << cfg.comment.ToUTF8();
	
	// Store body.
	beaconBody_ = body.str();
}



Beacon::~Beacon()
{
	while ( !objects_.empty() )
	{
		delete objects_.front().packet;
		objects_.pop_front();
	}
	
	while ( !beacons_.empty() )
	{
		delete beacons_.begin()->second.packet;
		beacons_.erase(beacons_.begin());
	}
}



wxThread::ExitCode Beacon::Entry()
{
	// Lock mutex and allow main to go waiting for it.
	mutex_.Lock();
	initialized_ = true;

	// Read objects from file.
	if ( !readObjectFile() )
	{
		keepgoing_ = false;
	}
	
	// Allow main to continue.
	mutex_.Unlock();
	
	// Keep beaconing.
	while ( keepGoing() )
	{
		// Check if next beaconing is at hand.
		mutex_.Lock();
		if ( !beacons_.empty() && time(NULL) >= beacons_.begin()->first )
		{
			// Create a copy of the packet to be transmitted.
			Packet* output = new Packet(*beacons_.begin()->second.packet);
			
			// Update timestamp of the copy to current moment.
			output->updateTimestamp();
			
			// Inform dupe control about the packet.
			dupefilter_->isDupe(output->getAprsPacket());
			
			// Transmit the packet.
			beacons_.begin()->second.con->write(output);
			
			// Re-schedule the beacon.
			beacons_.insert( make_pair(time(NULL)+beacons_.begin()->second.interval, beacons_.begin()->second) );
			
			// Unschedule this beacon.
			beacons_.erase(beacons_.begin());
		}
		mutex_.Unlock();

		// Need for beaconing is checked once a second.
		Sleep(1000);
	}
	
	return NULL;
}



void Beacon::waitForInit()
{
	while ( !initialized_ )
	{
		// Main thread does not need CPU now.
		Yield();
	}
	
	// Packet log thread is has started and is initializing. Let's wait for it to complete.
	mutex_.Lock();
	
	// Okay, initialization complete.
	mutex_.Unlock();
}



bool Beacon::operator!() const
{
	return !keepgoing_;
}



void Beacon::terminate()
{
	mutex_.Lock();
	keepgoing_ = false;
	mutex_.Unlock();
}



bool Beacon::addConnection(Connection* con, char const& conid, vector<wxString> const& path)
{
	// Dummy check.
	if ( con == NULL )
	{
		return false;
	}
	
	// Convert path to C++ string.
	vector<string> coolpath;
	for ( unsigned int i = 0; i < path.size(); i++ )
	{
		coolpath.push_back( static_cast<char const*>(path.at(i).ToAscii()) );
	}
	
	// Create beacon packet for this con.
	bool isAX25 = (conid != 'u');
	Packet* packet = new Packet('_', conid, new AprsPacket(owncall_, OWN_TARGET, coolpath, beaconBody_, isAX25), 'b');
	
	// Check packet.
	string errmsg;
	if ( packet->getAprsPacket()->error(errmsg) )
	{
		wxString msg = wxString::FromAscii(errmsg.c_str());
		wxLogError( wxString::Format(wxT("Failed to create beacon packet for connection %c: "), conid) + msg );
		wxLogDebug(wxString::FromAscii(packet->getAprsPacket()->getLogLine().c_str()));
		delete packet;
		return false;
	}
	
	// Create schedule for the connection.
	BeaconPacket bp;
	bp.packet = packet;
	bp.con = con;
	bp.interval = myInterval_;
	
	// Schedule beaconing to happen ASAP.
	mutex_.Lock();
	beacons_.insert( make_pair(0, bp) );
	mutex_.Unlock();
	
	
	// Schedule objects for this connection.
	for ( list<ObjectPacket>::const_iterator i = objects_.begin(); i != objects_.end(); i++ )
	{
		// Create copy of object packet for this con. If con is uplink, replace path.
		AprsPacket* tmpPacket = NULL;
		if ( conid == 'u' )
		{
			tmpPacket = new AprsPacket(owncall_, OWN_TARGET, coolpath, i->packet->getBody(), false);
		}
		else
		{
			tmpPacket = new AprsPacket(*i->packet);
		}
		
		Packet* packet = new Packet('_', conid, tmpPacket, 'o');
		
		BeaconPacket bp;
		bp.packet = packet;
		bp.con = con;
		bp.interval = i->interval;
		
		mutex_.Lock();
		beacons_.insert( make_pair(time(NULL)+i->offset, bp) );
		mutex_.Unlock();
	}
	
	return true;
}




bool Beacon::readObjectFile()
{
	// Check if we should read object file.
	if ( objFileName_.empty() )
	{
		return true;
	}
	
	// Add path to file name.
	string fullname = "";
#ifndef __WXMSW__
	fullname = string(SYSCONFDIR) + string("/");
#endif
	fullname += objFileName_;
		
	// Open file.
	wxLogMessage(wxT("Reading objects from ") + wxString::FromAscii(objFileName_.c_str()) + wxT("."));
	ifstream objfile (fullname.c_str());
	if ( !objfile )
	{
		wxLogError(wxT("Failed to open object file."));
		return false;
	}

	// Read all lines.
	string line;
	unsigned int lineno = 0;
	while ( getline(objfile, line) )
	{
		lineno++;
	
		// Check for too short or comment line.
		if ( line.length() < 31 || line.at(0) == '#' )
		{
			continue;
		}
		
		// Get offset.
		string offset = "";
		while ( !line.empty() && isdigit(line.at(0)) )
		{
			offset += line.at(0);
			line.erase(0,1);
		}
		
		// Skip space.
		if ( line.empty() || line.at(0) != ' ' )
		{
			wxLogError(wxString::Format(wxT("Line %d of object file is invalid."), lineno));
			return false;
		}
		line.erase(0,1);
		
		// Get interval.
		string interval = "";
		while ( !line.empty() && isdigit(line.at(0)) )
		{
			interval += line.at(0);
			line.erase(0,1);
		}
			
		// Skip space.
		if ( line.empty() || line.at(0) != ' ' )
		{
			wxLogError(wxString::Format(wxT("Line %d of object file is invalid."), lineno));
			return false;
		}
		line.erase(0,1);
		
		// Get path.
		vector<string> opath;
		string pElem = "";
		while ( !line.empty() && line.at(0) != ' ' )
		{
			if ( line.at(0) == ',' )
			{
				opath.push_back(pElem);
				pElem.clear();
			}
			else
			{
				pElem += line.at(0);
			}
			line.erase(0,1);
		}
		if ( !line.empty() && !pElem.empty() && pElem != "nopath" )
		{
			opath.push_back(pElem);
		}
		
		// Skip space.
		if ( line.empty() || line.at(0) != ' ' )
		{
			wxLogError(wxString::Format(wxT("Line %d of object file is invalid."), lineno));
			return false;
		}
		line.erase(0,1);
		
		// Rest on the line is the object.
		
		// Compile packet for the object.
		AprsPacket* tmpPacket = new AprsPacket(owncall_, OWN_TARGET, opath, line, true);
		string errmsg;
		if ( tmpPacket->error(errmsg) )
		{
			wxLogError( wxString::Format(wxT("Failed to create carrier packet for object file line %d: "), lineno) + wxString::FromAscii(errmsg.c_str()) );
			delete(tmpPacket);
			return false;
		}
		
		// Create packet, convert times to seconds while at it.
		ObjectPacket op;
		op.packet = tmpPacket;
		wxString::FromAscii(offset.c_str()).ToLong(&op.offset);
		op.offset *= 60;
		wxString::FromAscii(interval.c_str()).ToLong(&op.interval);
		op.interval *= 60;
		
		// Save packet.
		objects_.push_back(op);
		
	} // while lines
	
	// Reading done.
	objfile.close();
	return true;
}



bool Beacon::keepGoing()
{
	bool retval;
	mutex_.Lock();
	retval = keepgoing_;
	mutex_.Unlock();
	return retval;
}
