/* $Id: filterelement.cpp 427 2011-01-23 22:19:13Z oh2gve $
 *
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen 
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Own stuff.
#include "filterelement.h"
#include <fap.h>

// std stuff.
#include <vector>
using std::vector;
#include <string>
using std::string;

// wx stuff.
#include <wx/string.h>
#include <wx/regex.h>



FilterElement::FilterElement():
areas_(), ranges_(), buddyRegs_(), buddyPatterns_()
{
}

FilterElement::FilterElement(FilterElement const& old):
areas_(old.areas_), ranges_(old.ranges_), buddyRegs_(), buddyPatterns_(old.buddyPatterns_)
{
	for ( unsigned int i = 0; i < old.buddyRegs_.size(); i++ )
	{
		wxRegEx* tmp = new wxRegEx(old.buddyPatterns_.at(i));
		buddyRegs_.push_back(tmp);
	}
}

FilterElement::~FilterElement()
{
	for ( unsigned int i = 0; i < buddyRegs_.size(); i++ )
	{
		delete buddyRegs_.at(i);
	}
}


void FilterElement::addArea(Area const& a)
{
	areas_.push_back(a);
}

void FilterElement::addRange(Range const& r)
{
	ranges_.push_back(r);
}
	
void FilterElement::addBuddyPattern(wxString const& p)
{
	buddyPatterns_.push_back(p);
	buddyRegs_.push_back(new wxRegEx(p));
}



bool FilterElement::hasBuddyPatterns() const
{
	return !buddyRegs_.empty();
}

unsigned int FilterElement::getPartCount() const
{
	return areas_.size() + buddyRegs_.size() + ranges_.size();
}
	
bool FilterElement::accepts(string const& call, double const& lon, double const& lat) const
{
	// Convert callsign into wxString for matching.
	wxString callsign = wxString::FromAscii(call.c_str());
	
	// Check range.
	if ( !ranges_.empty() )
	{
		bool match = false;
		for ( unsigned int i = 0; i < ranges_.size(); ++i )
		{
			// If packet is in range, it's good.
			double range = fap_distance(lon, lat, ranges_.at(i).lon, ranges_.at(i).lat);
			if ( range <= ranges_.at(i).dist )
			{
				match = true;
				break;
			}
		}
		if ( !match )
		{
			return false;
		}
	}
	
	// Check callsign patterns for matches.
	if ( !buddyRegs_.empty() )
	{
		bool match = false;
		for ( unsigned int i = 0; i < buddyRegs_.size(); i++ )
		{
			if ( buddyRegs_.at(i)->Matches(callsign) )
			{
				match = true;
				break;
			}
		}
		if ( !match )
		{
			return false;
		}
	}
	
	// Check area.
	if ( !areas_.empty() )
	{
		bool match = false;
		for ( unsigned int i = 0; i < areas_.size(); ++i )
		{
			// If packet is in this area, it's going to pass.
			if ( lat <= areas_.at(i).latN && lat >= areas_.at(i).latS &&
				  lon >= areas_.at(i).lonW && lon <= areas_.at(i).lonE )
			{
				match = true;
				break;
			}
		}
		if ( !match )
		{
			return false;
		}
	}
		return true;
}


bool FilterElement::accepts(string const& call) const
{
	// Convert callsign into wxString for matching.
	wxString callsign = wxString::FromAscii(call.c_str());
	
	// If range or area parts are available, we fail straight away.
	if ( !ranges_.empty() || !areas_.empty() )
	{
		return false;
	}
	
	// Check callsign patterns for matches.
	if ( !buddyRegs_.empty() )
	{
		bool match = false;
		for ( unsigned int i = 0; i < buddyRegs_.size(); i++ )
		{
			if ( buddyRegs_.at(i)->Matches(callsign) )
			{
				match = true;
				break;
			}
		}
		if ( !match )
		{
			return false;
		}
	}
		return true;
}
