/* $Id: logserver.cpp 428 2011-01-26 08:03:42Z oh2gve $
 *
 * Copyright 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// own stuff.
#include "logserver.h"

// std stuff.
#include <string>
using std::string;
#include <map>
using std::map;
using std::make_pair;
#include <ctime>
#include <sstream>
using std::ostringstream;

// wx stuff.
#include <wx/string.h>
#include <wx/log.h>
#include <wx/datetime.h>
#include <wx/socket.h>


LogServer::LogServer(wxIPV4address const& cfg, wxString const& ownCall):
initialized_(false), ownCall_(ownCall), masterAddr_(cfg), masterSocket_(NULL), clients_(), lastHeartBeat_(time(NULL))
{
	// Create master socket. We want non-blocking I/O.
	masterSocket_ = new wxSocketServer(masterAddr_, wxSOCKET_NOWAIT | wxSOCKET_REUSEADDR);

	// Check how master socket creation went.
	if ( masterSocket_ != NULL && masterSocket_->Ok() )
	{
		initialized_ = true;
		wxLogMessage(wxT(" logserver: Listening for connections at ") + masterAddr_.Hostname() + wxT(":") + wxString::Format(wxT("%d."), masterAddr_.Service()) );
	}
	else
	{
		wxLogError(wxT(" logserver: Initialization failed."));
	}
}


LogServer::~LogServer()
{
	// Disconnect client sockets.
	for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
	{
		i->second->Destroy();
		wxLogMessage(wxT(" logserver: disconnected ") + i->first + wxT("."));
	}
	
	// Remove master socket.
	if ( masterSocket_ != NULL )
	{
		masterSocket_->Destroy();
	}
}


bool LogServer::operator!()
{
	return !initialized_;
}


void LogServer::accept()
{
	// Check socket for incoming connections.
	wxSocketBase* newClient;
	if ( (newClient = masterSocket_->Accept(false)) != NULL )
	{
		wxLogMessage(wxT(" logserver: Incoming connection detected."));
		
		// We want non-blocking I/O.
		newClient->SetFlags(wxSOCKET_NOWAIT);
		
		// Get peer data.
		wxIPV4address clientAddr;
		wxString clientAddrStr;
		if ( newClient->GetPeer(clientAddr) )
		{
			clientAddrStr = clientAddr.Hostname();
			clientAddrStr += wxT(":");
			clientAddrStr << clientAddr.Service();
			wxLogMessage(wxT(" logserver: Connetion from ") + clientAddrStr + wxT(" accepted."));
		}
		else
		{
			clientAddrStr = wxT("unknown");
			clientAddrStr << time(NULL);
			wxLogMessage(wxT(" logserver: Can't get figure out source ip or FQDN of the incoming connection, but accepting it anyway."));
		}
		
		// Save connection.
		clients_.insert(make_pair<wxString, wxSocketBase*>(clientAddrStr, newClient));
		
		// Greet the log watcher.
		ostringstream greet;
		greet << "# Welcome to " << string(ownCall_.ToAscii()) << "'s packet monitor port.\r\n";
		greet << "# Hit q and enter when you want to quit.\r\n";
		newClient->Write(greet.str().c_str(), greet.str().size());
		
		// Check that correct amount of bytes were written.
		if ( newClient->LastCount() != greet.str().size() )
		{
			wxString num1;
			num1 << newClient->LastCount();
			wxString num2;
			num2 << greet.str().size();
			wxLogDebug(wxT(" logserver: Only ") + num1 + wxT(" bytes of ") + num2 + wxT(" intended could be written to ") + clientAddrStr + wxT("."));
		}
	}
	
	// Check if our clients are still connected.
	for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
	{
		// Check for disconnected/failed ones. Check input from connected ones.
		if ( i->second->IsDisconnected() )
		{
			if ( i->second->Error() )
			{
				logError(i->second->LastError());
			}
			wxLogMessage(wxT(" logserver: It seems that connection from ") + i->first + wxT(" has been disconnected. Closing it."));
		}
		else
		{
			// Check if user has sent something for us.
			char byte = 0x00;
			i->second->Read(&byte, 1);
			if ( i->second->LastCount() == 1 )
			{
				// Check if we have received a 'q'.
				if ( byte != 'q' )
				{
					// No 'q' from this connection, we continue with others.
					continue;
				}
				// We got 'q'.
				string msg = "Have a nice day!\r\n";
				i->second->Write(msg.c_str(), msg.size());
				wxLogMessage(wxT(" logserver: Connection from ") + i->first + wxT(" was closed by the client."));
			}
			else
			{
				// No input from the user.
				continue;
			}
		}
		
		// Okay let's close the connection.
		i->second->Destroy();
		clients_.erase(i);
		
		// i is now invalid, so don't check any more connections this time.
		break;
	}
	
	// Check if we shoud heartbeat to the connections.
	if ( lastHeartBeat_ + 15 < time(NULL) )
	{
		ostringstream hbline;
		wxDateTime timestamp = wxDateTime::Now();
		hbline << "# " << timestamp.Format(wxT("%Y-%m-%d %H:%M:%S"), wxDateTime::UTC).ToAscii() << "\r\n";
		for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
		{
			// Send the line.
			i->second->Write(hbline.str().c_str(), hbline.str().size());
			// Check that correct amount of bytes were written.
			if ( i->second->LastCount() != hbline.str().size() )
			{
				wxString num1;
				num1 << i->second->LastCount();
				wxString num2;
				num2 << hbline.str().size();
				wxString num3;
				num3 << i->first;
				wxLogDebug(wxT(" logserver: Only ") + num1 + wxT(" bytes of ") + num2 + wxT(" intended could be written to connection ") + num3 + wxT("."));
			}
		}
		lastHeartBeat_ = time(NULL);
	}
}


void LogServer::write(string const& line)
{
	// Send the line to them.
	for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
	{
		// Send the line.
		i->second->Write(line.c_str(), line.size());
		// Check that correct amount of bytes were written.
		if ( i->second->LastCount() != line.size() )
		{
			wxString num1;
			num1 << i->second->LastCount();
			wxString num2;
			num2 << line.size();
			wxString num3;
			num3 << i->first;
			wxLogDebug(wxT(" logserver: Only ") + num1 + wxT(" bytes of ") + num2 + wxT(" intended could be written to ") + num3 + wxT("."));
		}
		
	}
}


void LogServer::logError(wxSocketError const& lastError) const
{
	wxString error;

	switch (lastError) {
	  case wxSOCKET_NOERROR:
		error = wxT("No error happened.");
		break;
	  case wxSOCKET_INVOP:
		error = wxT("Invalid operation.");
		break;
	  case wxSOCKET_IOERR:
		error = wxT("Input/Output error.");
		break;
	  case wxSOCKET_INVADDR:
		error = wxT("Invalid address passed to wxSocket.");
		break;
	  case wxSOCKET_INVSOCK:
		error = wxT("Invalid socket (uninitialized).");
		break;
	  case wxSOCKET_NOHOST:
		error = wxT("No corresponding host.");
		break;
	  case wxSOCKET_INVPORT:
		error = wxT("Invalid port.");
		break;
	  case wxSOCKET_WOULDBLOCK:
		error = wxT("The socket is non-blocking and the operation would block.");
		break;
	  case wxSOCKET_TIMEDOUT:
		error = wxT("The timeout for this operation expired.");
		break;
	  case wxSOCKET_MEMERR:
		error = wxT("Memory exhausted.");
		break;

	  default:
		error = wxT("Error 12765. WTF???");
	}

	wxString msg = wxT(" logserver: ");
	msg += error;
	wxLogError(msg);
}
