/* $Id$
 *
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

/**
 * \mainpage
 * \section modules Modules
 * Main module is one big chunck containg almost all things. In addition
 * there's one smaller module, Config, which contains configuration file parser
 * and storage for cfg values.
 * \section threads Threads
 * aprsg has a main thread, and one thread for each connection. There is also three assistant threads, which handle packet logging, duplicate detection and beaconing.
 */

#ifndef MAIN_H
#define MAIN_H

// Own stuff.
#ifdef __WXMSW__
#	include "windows/window.h"
#	include "windows/controller.h"
#endif
#include "cfg.h"
#include "connection.h"
#include "uplink.h"
#include "packetlog.h"
#include "dupefilter.h"
#include "beacon.h"

// Std stuff.
#include <fstream>

// Wx stuff.
#include <wx/app.h>
#include <wx/string.h>
#include <wx/file.h>


/// Main app class.
/**
 * Used to attach our code into wxwidgets framework.
 */
#ifdef __WXMSW__
class AprsApp: public wxApp
#else
class AprsApp: public wxAppConsole
#endif
{
public:

	// The usual ones.
	AprsApp();
	virtual ~AprsApp();

	/// Initializations.
	virtual bool OnInit();

	// Cleanup.
	virtual int OnExit();

#ifndef __WXMSW__
	/// Command line parameter initialization.
	virtual void OnInitCmdLine(wxCmdLineParser& parser);

	/// Framework gives us the command line parameters via this.
	virtual bool OnCmdLineParsed(wxCmdLineParser& parser);

	/// Receives signal number from our OS signal handler.
	void gotSignal(int num);

	// Normal operation loop.
	virtual int OnRun();
#endif

private:

#ifdef __WXMSW__
	// Status window.
	Window* window_;
	
	// Controller thread does the actual gating action.
	Controller* controller_;
#endif

	/// Opens main log file and sets it as logging target.
	bool openLog(bool const initialOpening = true);

	/// If main log file is open, closes it and resets logging target.
	void closeLog();
	
	/// Starts packet logging and dupe filter.
	bool startAssistants();
	
	/// Stops packet logging and dupe filter.
	void stopAssistants();

	/// Opens connections based on configuration_.
	bool openConnections();

	/// Close all open connections.
	void closeConnections();
	
#ifndef __WXMSW__	
	/// Flag for telling difference between foreground (to be shut down)
	/// and backround processes when switching from OnInit() to OnRun().
	int iAmDaddy_;

	/// Flag for delievering signal reception info into main loop.
	int gotSignal_;
	
	/// File descriptor of /dev/null.
	int nullFd_;
	
	/// Flag for daemon or foreground mode.
	bool daemonMode_;
	
	/// Create us a pid file.
	bool createPidFile();

	/// Remove our pid file.
	void removePidFile();

	/// Pid file handle.
	wxFile* pidFile_;

	/// Reopen general log.
	bool reopenLog();


	/// Configuration file name.
	/**
	 * Can be given from command line.
	 */
	wxString configFileName_;
#endif

	/// Main log file. NULL if log goes to to stdout.
	std::ofstream* logFile_;
	
	/// Packet logger.
	Packetlog* packetlog_;
	
	/// Duplicate filter.
	DupeFilter* dupefilter_;
	
	/// Beacon manager.
	Beacon* beacon_;

	/// IGate configuration.
	Cfg::Configuration configuration_;

	/// Server connection.
	Uplink* uplink_;

	/// Connection storage.
	std::vector<Connection*> connections_;
};


#endif // MAIN_H
