/* $Id: uplink.cpp 453 2011-02-27 10:06:27Z oh2gve $
 *
 * Copyright 2007 Antti Seppänen
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Own stuff.
#include "uplink.h"
#include "cfg.h"
using Cfg::UplinkCfg;
#include "packetlog.h"
#include "dupefilter.h"
#include "aprspacket.h"

// std stuff.
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <ctime>

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>



Uplink::Uplink(UplinkCfg const& cfg, string const& login, Packetlog* const& packetlog, DupeFilter* const& dupefilter):
Connection(cfg.serverState, 'u', packetlog, dupefilter),
socket_(NULL), packetlog_(packetlog), cfg_(cfg), login_(login),
currentServer_(), currentServerNo_(0), lastTrafficTime_(time(NULL)), lastTryTime_(0),
inputBuffer_()
{
	// Save current server as string.
	currentServer_ = cfg_.serverNames.at(currentServerNo_);
}


Uplink::~Uplink()
{
	if (socket_ != 0)
	{
	  socket_->Destroy();
	}
}


void* Uplink::Entry()
{
	// Tell main that we're alive.
	threadStarted();
	
	// Try to connect.
	if ( !connect() )
	{
		wxLogMessage(wxT(" uplink: Initial connection attempt failed. Let's wait a little and try again."));
	}
	startupDone();
	
	// Main loop of this thread.
	while ( keepGoing() )
	{
		// Either operate on connection or try to fix it.
		if ( socketIsConnected() )
		{
			// Read as much data from the socket buffer as possible but not more than 8 kB.
			char buf[8192];
			socket_->Read(buf, WXSIZEOF(buf));

			// If we received bytes, mark this as the latest moment we got data from the socket.
			if ( socket_->LastCount() > 0 )
			{
				lastTrafficTime_ = time(NULL);
			}
				
			// Save received bytes.

			// Append data from buf to storage string character by character.
			bool gotCR = false;
			for (unsigned int i = 0; i < socket_->LastCount(); i++)
			{
				// Check for CR.
				if ( buf[i] == '\r')
				{
					gotCR = true;
				}
				else if ( gotCR && buf[i] == '\n' )
				{
					// Remove CR.
					inputBuffer_.erase(inputBuffer_.size()-1);
					
					// If line begins with #, we discard it.
					if ( !inputBuffer_.empty() && inputBuffer_.at(0) == '#' )
					{
						inputBuffer_.clear();
						continue;
					}

					// Send packet to main.
					inputPush(AprsPacket::fromTnc2Packet(inputBuffer_));
					
					// Clear the buffer for next packet.
					inputBuffer_.clear();

					// No need to store the \n.
					continue;
				}

				// Store the char.
				inputBuffer_.append(1, buf[i]);
			}
			
			// Then let's do tx to the internet.
			Packet* output = outputPop();

			// If we got something, write it.
			if ( output != NULL )
			{
				// Get the data to be written.
				string packet = output->getAprsPacket()->getTnc2Packet() + "\r\n";

				// Write it.
				socket_->Write(packet.c_str(), packet.size());
				if ( socket_->LastCount() != packet.size() )
				{
					wxLogMessage(wxString::Format(wxT(" uplink: Warning! Tried to send packet (%d bytes), but socket says it sent %d bytes."), packet.size(), socket_->LastCount()));
				}
				
				// That's it, delete the packet.
				packetlog_->write(output);
			}
			else
			{
				delete output;
			}
		}
		else // socketIsConnected() returned false
		{
			// Try to reconnect.
			if ( reconnect() )
			{
				wxLogMessage(wxT(" uplink: connection reopened, normal operation restarted."));
			}
			else
			{
				wxLogMessage(wxT(" uplink: unable to recover from error, giving up."));
				weHasError();
				break;
			}
		} // if (socketIsConnected())
		
		Sleep(10);
	} // while ()
	
	
	// When we get here the socket is no longer connected. Destroy it.
	socket_->Destroy();
	socket_ = NULL;

	// Terminate this thread.
	return NULL;
}


bool Uplink::connect()
{
	// Tell user what we're going to do.
	wxLogMessage(wxT(" uplink: trying to connect to ") + currentServer_ + wxT("..."));
	
	// If old socket exists, clear it away.
	if (socket_ != NULL)
	{
		socket_->Destroy();
	}
	
	// Create a new socket client with non-blocking I/O and without events.
	socket_ = new wxSocketClient(wxSOCKET_NOWAIT);
	socket_->Notify(false);

	// Connect to address and wait 10 seconds for connection.
	socket_->Connect(cfg_.servers.at(currentServerNo_), false);
	socket_->WaitOnConnect(10);
	
	// Check for failures in connection.
	if ( !socket_->IsConnected() || socket_->Error() )
	{
		wxLogMessage(wxT(" uplink: failed."));
		return false;
	}
	
	// Do login.
	socket_->Write(login_.c_str(), login_.size());
	if ( socket_->LastCount() != login_.size() )
	{
		wxLogMessage(wxString::Format(wxT(" uplink: Warning! Tried to send login (%d bytes), but socket says it sent %d bytes."), login_.size(), socket_->LastCount()));
	}
	
	// Tell user that we succeeded.
	wxLogMessage(wxT(" uplink: connected successfully! "));
	
	return true;
}


bool Uplink::socketIsConnected()
{
	// Test if the socket exists.
	if ( socket_ == NULL )
	{
		wxLogMessage(wxT(" uplink: socket has disappeared. (wtf?)"));
		return false;
	}

	// Test if the socket is connected.
	if ( !socket_->IsConnected() )
	{
		logError();
		return false;
	}

	// If enabled, check that we have received data recently.
	if ( cfg_.noTrafficTimeout > 0 )
	{
		if ( time(NULL) - lastTrafficTime_ > cfg_.noTrafficTimeout*60 )
		{
			wxString msg = wxT(" uplink: No traffic received from ");
			msg += currentServer_;
			msg += wxT(" during last ");
			msg << cfg_.noTrafficTimeout;
			msg += wxT(" minutes, seems that connection has died.");
			
			wxLogMessage(msg);
			return false;
		}
	}
			
	return true;
}


bool Uplink::reconnect()
{
	// If we're having an i/o-error or some crazy blocking error, it might be that the connection
	// has just timed out. So let's try to reconnect.

	if ( socket_->LastError() != wxSOCKET_IOERR &&
		  socket_->LastError() != wxSOCKET_WOULDBLOCK )
	{
		return false;
	}
		
	// Keep trying until connected or shutdown request comes.
	while ( keepGoing() )
	{
		// If we have tried to connect recently, wait a little before trying again.
		if ( lastTryTime_ + cfg_.retryInterval*60 >= time(NULL) )
		{
			wxString msg = wxT(" uplink: waiting ");
			msg << (lastTryTime_ + cfg_.retryInterval - time(NULL));
			msg += wxT(" minutes before trying again.");
			wxLogMessage(msg);
		}
		while ( lastTryTime_ + cfg_.retryInterval >= time(NULL) )
		{
			if ( !keepGoing() )
			{
				return false;
			}
			Sleep(1000);
		}
		
		// Try to connect to all servers.
		for ( unsigned int i = 0; i < cfg_.servers.size(); i++ )
		{
			// Save server number.
			currentServerNo_ = i;
			
			// Save the chosen server into the string.
			currentServer_ = cfg_.serverNames.at(currentServerNo_);
			
			// Flush outgoing packets to avoid delayed gating.
			Packet const* output = outputPop();
			while ( output != NULL )
			{
				delete output;
				output = outputPop();
			}
				
			// Try to connect.
			if ( connect() )
			{
				// We're now successfully connected, which means that we can to try to
				// reconnect sooner than after retryInterval (if required).
				lastTryTime_ = 0;
				
				// Last info from the server has just been received, so we can reset that
				// timer too.
				lastTrafficTime_ = time(NULL);
			
				return true;
			}
		}
		
		// No luck this time.
		lastTryTime_ = time(NULL);
		
		// Check that our error has not changed to an unhandleable one.
		if ( socket_->LastError() != wxSOCKET_IOERR &&
			  socket_->LastError() != wxSOCKET_WOULDBLOCK )
		{
			return false;
		}
	}
	
	return false;
}
	
	
void Uplink::logError() const
{
	wxSocketError lastError = socket_->LastError();
	wxString error;

	switch (lastError) {
	  case wxSOCKET_NOERROR:
		error = wxT("No error happened.");
		break;
	  case wxSOCKET_INVOP:
		error = wxT("Invalid operation.");
		break;
	  case wxSOCKET_IOERR:
		error = wxT("Input/Output error.");
		break;
	  case wxSOCKET_INVADDR:
		error = wxT("Invalid address passed to wxSocket.");
		break;
	  case wxSOCKET_INVSOCK:
		error = wxT("Invalid socket (uninitialized).");
		break;
	  case wxSOCKET_NOHOST:
		error = wxT("No corresponding host.");
		break;
	  case wxSOCKET_INVPORT:
		error = wxT("Invalid port.");
		break;
	  case wxSOCKET_WOULDBLOCK:
		error = wxT("The socket is non-blocking and the operation would block.");
		break;
	  case wxSOCKET_TIMEDOUT:
		error = wxT("The timeout for this operation expired.");
		break;
	  case wxSOCKET_MEMERR:
		error = wxT("Memory exhausted.");
		break;

	  default:
		error = wxT("Error 12765. WTF???");
	}

	wxString msg = wxT(" uplink: connection to ");
	msg += currentServer_;
	msg += wxT(" failed: ");
	msg += error;
	
	wxLogMessage(msg);
}
